@@ -22,6 +22,9 @@
 #include <limits.h>
 #include <stdlib.h>
 #include <unistd.h>
+#if defined(_POSIX_PRIORITY_SCHEDULING) && defined(HAVE_SCHED_YIELD)
+#include <sched.h>
+#endif
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
@@ -77,9 +80,12 @@ gid_t GdmGroupId;		/* Groupid under which gdm should run */
 pid_t extra_process = 0;	/* An extra process.  Used for quickie 
 				   processes, so that they also get whacked */
 int extra_status = 0;		/* Last status from the last extra process */
-pid_t gdm_main_pid = 0;
+pid_t gdm_main_pid = 0;		/* PID of the main daemon */
 gboolean preserve_ld_vars = FALSE; /* Preserve the ld environment variables */
 gboolean no_daemon = FALSE;	/* Do not daemonize */
+gboolean no_console = FALSE;	/* There are no local servers, this means,
+				   don't run local servers and second,
+				   don't display info on the console */
 
 GdmConnection *fifoconn = NULL; /* Fifo connection */
 GdmConnection *pipeconn = NULL; /* slavepipe (handled just like Fifo for compatibility) connection */
@@ -101,6 +107,7 @@ gboolean gdm_emergency_server = FALSE;
 gboolean gdm_first_login = TRUE;
 
 int gdm_in_signal = 0;
+gboolean gdm_in_final_cleanup = FALSE;
 
 /* Configuration options */
 gchar *GdmUser = NULL;
@@ -212,6 +219,58 @@ compare_displays (gconstpointer a, gconstpointer b)
 		return 0;
 }	
 
+static void
+check_servauthdir (struct stat *statbuf)
+{
+    int r;
+
+    /* Enter paranoia mode */
+    IGNORE_EINTR (r = stat (GdmServAuthDir, statbuf));
+    if (r < 0) {
+	    char *s = g_strdup_printf
+		    (_("Server Authorization directory "
+		       "(daemon/ServAuthDir) is set to %s "
+		       "but this does not exist. Please "
+		       "correct gdm configuration %s and "
+		       "restart gdm."), GdmServAuthDir,
+		     GDM_CONFIG_FILE);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
+	    gdm_fail (_("%s: Authdir %s does not exist. Aborting."), "gdm_config_parse", GdmServAuthDir);
+    }
+
+    if (! S_ISDIR (statbuf->st_mode)) {
+	    char *s = g_strdup_printf
+		    (_("Server Authorization directory "
+		       "(daemon/ServAuthDir) is set to %s "
+		       "but this is not a directory. Please "
+		       "correct gdm configuration %s and "
+		       "restart gdm."), GdmServAuthDir,
+		     GDM_CONFIG_FILE);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
+	    GdmPidFile = NULL;
+	    gdm_fail (_("%s: Authdir %s is not a directory. Aborting."), "gdm_config_parse", GdmServAuthDir);
+    }
+}
+
+static void
+check_logdir (void)
+{
+	struct stat statbuf;
+	int r;
+   
+	IGNORE_EINTR (r = stat (GdmLogDir, &statbuf));
+	if (r < 0 ||
+	    ! S_ISDIR (statbuf.st_mode))  {
+		gdm_error (_("%s: Logdir %s does not exist or isn't a directory.  Using ServAuthDir %s."), "gdm_config_parse",
+			   GdmLogDir, GdmServAuthDir);
+		g_free (GdmLogDir);
+		GdmLogDir = g_strdup (GdmServAuthDir);
+	}
+}
+
 /**
  * gdm_config_parse:
  *
@@ -227,11 +286,13 @@ gdm_config_parse (void)
     gchar *bin;
     VeConfig *cfg;
     GList *list, *li;
+    int r;
     
     displays = NULL;
     high_display_num = 0;
 
-    if (stat (GDM_CONFIG_FILE, &statbuf) == -1) {
+    IGNORE_EINTR (r = stat (GDM_CONFIG_FILE, &statbuf));
+    if (r < 0) {
 	    gdm_error (_("%s: No configuration file: %s. Using defaults."),
 		       "gdm_config_parse", GDM_CONFIG_FILE);
     } else {
@@ -393,16 +454,6 @@ gdm_config_parse (void)
 	    gdm_error (_("%s: No remote greeter specified."), "gdm_config_parse");
     }
 
-    if (ve_string_empty (GdmServAuthDir)) {
-	    gdm_text_message_dialog
-		    (_("No daemon/ServAuthDir specified in the configuration file"));
-	    GdmPidFile = NULL;
-	    gdm_fail (_("%s: No authdir specified."), "gdm_config_parse");
-    }
-
-    if (ve_string_empty (GdmLogDir))
-	GdmLogDir = GdmServAuthDir;
-
     if (ve_string_empty (GdmSessDir))
 	gdm_error (_("%s: No sessions directory specified."), "gdm_config_parse");
 
@@ -457,7 +508,9 @@ gdm_config_parse (void)
 
     /* Find local X server definitions */
     list = ve_config_get_keys (cfg, GDM_KEY_SERVERS);
-    for (li = list; li != NULL; li = li->next) {
+    /* only read the list if no_console is FALSE
+       at this stage */
+    for (li = list; ! no_console && li != NULL; li = li->next) {
 	    const char *key = li->data;
 	    if (isdigit (*key)) {
 		    char *full;
@@ -496,6 +549,13 @@ gdm_config_parse (void)
 
     if (displays == NULL && ! GdmXdmcp) {
 	    char *server = NULL;
+
+	    /* if we requested no local servers (there is no console),
+	       then don't display errors in console messages */
+	    if (no_console) {
+		    gdm_fail (_("%s: XDMCP disabled and no local servers defined. Aborting!"), "gdm_config_parse");
+	    }
+
 	    if (access (GdmStandardXServer, X_OK) == 0) {
 		    server = GdmStandardXServer;
 	    } else if (access ("/usr/bin/X11/X", X_OK) == 0) {
@@ -536,6 +596,11 @@ gdm_config_parse (void)
 	    }
     }
 
+    /* If no displays were found, then obviously
+       we're in a no console mode */
+    if (displays == NULL)
+	    no_console = TRUE;
+
     /* Lookup user and groupid for the gdm user */
     pwent = getpwnam (GdmUser);
 
@@ -552,7 +617,8 @@ gdm_config_parse (void)
 		       "Please correct gdm configuration %s "
 		       "and restart gdm."),
 		     GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
 	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Can't find the gdm user (%s). Aborting!"), "gdm_config_parse", GdmUser);
     } else {
@@ -566,7 +632,8 @@ gdm_config_parse (void)
 		       "pose a security risk.  Please "
 		       "correct gdm configuration %s and "
 		       "restart gdm."), GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
 	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: The gdm user should not be root. Aborting!"), "gdm_config_parse");
     }
@@ -586,7 +653,8 @@ gdm_config_parse (void)
 		       "Please correct gdm configuration %s "
 		       "and restart gdm."),
 		     GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
 	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Can't find the gdm group (%s). Aborting!"), "gdm_config_parse", GdmGroup);
     } else  {
@@ -600,7 +668,8 @@ gdm_config_parse (void)
 		       "pose a security risk. Please "
 		       "correct gdm configuration %s and "
 		       "restart gdm."), GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
 	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: The gdm group should not be root. Aborting!"), "gdm_config_parse");
     }
@@ -640,35 +709,37 @@ gdm_config_parse (void)
     
     g_free (bin);
 
+    /* Check the serv auth and log dirs */
+    if (ve_string_empty (GdmServAuthDir)) {
+	    if ( ! no_console)
+		    gdm_text_message_dialog
+			    (_("No daemon/ServAuthDir specified in the configuration file"));
+	    GdmPidFile = NULL;
+	    gdm_fail (_("%s: No daemon/ServAuthDir specified."), "gdm_config_parse");
+    }
+
+    if (ve_string_empty (GdmLogDir)) {
+	g_free (GdmLogDir);
+	GdmLogDir = g_strdup (GdmServAuthDir);
+    }
 
     /* Enter paranoia mode */
-    if (stat (GdmServAuthDir, &statbuf) == -1)  {
-	    char *s = g_strdup_printf
-		    (_("Server Authorization directory "
-		       "(daemon/ServAuthDir) is set to %s "
-		       "but this does not exist. Please "
-		       "correct gdm configuration %s and "
-		       "restart gdm."), GdmServAuthDir,
-		     GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
-	    GdmPidFile = NULL;
-	    gdm_fail (_("%s: Authdir %s does not exist. Aborting."), "gdm_config_parse", GdmServAuthDir);
-    }
+    check_servauthdir (&statbuf);
 
-    if (! S_ISDIR (statbuf.st_mode)) {
-	    char *s = g_strdup_printf
-		    (_("Server Authorization directory "
-		       "(daemon/ServAuthDir) is set to %s "
-		       "but this is not a directory. Please "
-		       "correct gdm configuration %s and "
-		       "restart gdm."), GdmServAuthDir,
-		     GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
-	    GdmPidFile = NULL;
-	    gdm_fail (_("%s: Authdir %s is not a directory. Aborting."), "gdm_config_parse", GdmServAuthDir);
-    }
+    seteuid (0);
+    setegid (0);
 
-    if (statbuf.st_uid != GdmUserId || statbuf.st_gid != GdmGroupId)  {
+    /* Now set things up for us as  */
+    chown (GdmServAuthDir, 0, GdmGroupId);
+    chmod (GdmServAuthDir, (S_IRWXU|S_IRWXG|S_ISVTX));
+
+    setegid (GdmGroupId);
+    seteuid (GdmUserId);
+
+    /* again paranoid */
+    check_servauthdir (&statbuf);
+
+    if (statbuf.st_uid != 0 || statbuf.st_gid != GdmGroupId)  {
 	    char *s = g_strdup_printf
 		    (_("Server Authorization directory "
 		       "(daemon/ServAuthDir) is set to %s "
@@ -678,31 +749,35 @@ gdm_config_parse (void)
 		       "gdm."),
 		     GdmServAuthDir, GdmUser, GdmGroup,
 		     GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
 	    GdmPidFile = NULL;
 	    gdm_fail (_("%s: Authdir %s is not owned by user %s, group %s. Aborting."), "gdm_config_parse", 
-		      GdmServAuthDir, GdmUser, GdmGroup);
+		      GdmServAuthDir, gdm_root_user (), GdmGroup);
     }
 
-    if (statbuf.st_mode != (S_IFDIR|S_IRWXU|S_IRGRP|S_IXGRP))  {
+    if (statbuf.st_mode != (S_IFDIR|S_IRWXU|S_IRWXG|S_ISVTX))  {
 	    char *s = g_strdup_printf
 		    (_("Server Authorization directory "
 		       "(daemon/ServAuthDir) is set to %s "
 		       "but has the wrong permissions, it "
-		       "should have permissions of 0750. "
+		       "should have permissions of %o. "
 		       "Please correct the permissions or "
 		       "the gdm configuration %s and "
 		       "restart gdm."),
-		     GdmServAuthDir, GDM_CONFIG_FILE);
-	    gdm_text_message_dialog (s);
+		     GdmServAuthDir, (S_IRWXU|S_IRWXG|S_ISVTX), GDM_CONFIG_FILE);
+	    if ( ! no_console)
+		    gdm_text_message_dialog (s);
 	    GdmPidFile = NULL;
-	    gdm_fail (_("%s: Authdir %s has wrong permissions %o. Should be 0750. Aborting."), "gdm_config_parse", 
-		      GdmServAuthDir, statbuf.st_mode);
+	    gdm_fail (_("%s: Authdir %s has wrong permissions %o. Should be %o. Aborting."), "gdm_config_parse", 
+		      GdmServAuthDir, statbuf.st_mode, (S_IRWXU|S_IRWXG|S_ISVTX));
     }
 
     seteuid (0);
     setegid (0);
 
+    check_logdir ();
+
     /* Check that user authentication is properly configured */
     gdm_verify_check ();
 
@@ -751,6 +826,7 @@ gdm_daemonify (void)
 
         exit (EXIT_SUCCESS);
     }
+    gdm_main_pid = getpid ();
 
     if (pid < 0) 
 	gdm_fail (_("%s: fork() failed!"), "gdm_daemonify");
@@ -759,12 +835,12 @@ gdm_daemonify (void)
 	gdm_fail (_("%s: setsid() failed: %s!"), "gdm_daemonify",
 		  strerror(errno));
 
-    chdir (GdmServAuthDir);
+    IGNORE_EINTR (chdir (GdmServAuthDir));
     umask (022);
 
-    close (0);
-    close (1);
-    close (2);
+    IGNORE_EINTR (close (0));
+    IGNORE_EINTR (close (1));
+    IGNORE_EINTR (close (2));
 
     gdm_open_dev_null (O_RDONLY); /* open stdin - fd 0 */
     gdm_open_dev_null (O_RDWR); /* open stdout - fd 1 */
@@ -811,10 +887,11 @@ void
 gdm_final_cleanup (void)
 {
 	GSList *list, *li;
+	gboolean first;
 
 	gdm_debug ("gdm_final_cleanup");
 
-	gdm_sigchld_block_push ();
+	gdm_in_final_cleanup = TRUE;
 
 	if (extra_process > 1) {
 		/* we sigterm extra processes, and we
@@ -823,7 +900,47 @@ gdm_final_cleanup (void)
 		extra_process = 0;
 	}
 
-	gdm_sigchld_block_pop ();
+	/* First off whack all XDMCP and FLEXI_XNEST
+	   slaves, we'll wait for them later */
+	for (li = displays; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		if (d->type == TYPE_XDMCP ||
+		    d->type == TYPE_FLEXI_XNEST) {
+			/* set to DEAD so that we won't kill it again */
+			d->dispstat = DISPLAY_DEAD;
+			if (d->slavepid > 1)
+				kill (d->slavepid, SIGTERM);
+		}
+	}
+
+	/* Now completely unmanage the local servers */
+	first = TRUE;
+	list = g_slist_copy (displays);
+	/* somewhat of a hack to kill last server
+	 * started first.  This mostly makes things end up on
+	 * the right vt */
+	list = g_slist_reverse (list);
+	for (li = list; li != NULL; li = li->next) {
+		GdmDisplay *d = li->data;
+		if (d->type == TYPE_XDMCP ||
+		    d->type == TYPE_FLEXI_XNEST)
+			continue;
+		/* HACK! Wait 2 seconds between killing of local servers
+		 * because X is stupid and full of races and will hang my
+		 * keyboard if I don't */
+		if ( ! first) {
+			/* there could be signals happening
+			   here */
+			gdm_sleep_no_signal (2);
+		}
+		first = FALSE;
+		gdm_display_unmanage (d);
+	}
+	g_slist_free (list);
+
+	/* and now kill and wait for the XDMCP and FLEXI_XNEST
+	   slaves.  unmanage will not kill slaves we have already
+	   killed unless a SIGTERM was sent in the meantime */
 
 	list = g_slist_copy (displays);
 	for (li = list; li != NULL; li = li->next) {
@@ -836,21 +953,6 @@ gdm_final_cleanup (void)
 	}
 	g_slist_free (list);
 
-	list = g_slist_copy (displays);
-	/* somewhat of a hack to kill last server
-	 * started first.  This mostly makes things end up on
-	 * the right vt */
-	list = g_slist_reverse (list);
-	for (li = list; li != NULL; li = li->next) {
-		GdmDisplay *d = li->data;
-		/* HACK! Wait 2 seconds between killing of local servers
-		 * because X is stupid and full of races and will hang my
-		 * keyboard if I don't */
-		if (li != list)
-			sleep (2);
-		gdm_display_unmanage (d);
-	}
-	g_slist_free (list);
 
 	/* Close stuff */
 
@@ -861,7 +963,7 @@ gdm_final_cleanup (void)
 		char *path;
 		gdm_connection_close (fifoconn);
 		path = g_strconcat (GdmServAuthDir, "/.gdmfifo", NULL);
-		unlink (path);
+		IGNORE_EINTR (unlink (path));
 		g_free (path);
 		fifoconn = NULL;
 	}
@@ -872,20 +974,20 @@ gdm_final_cleanup (void)
 	}
 
 	if (slave_fifo_pipe_fd >= 0) {
-		close (slave_fifo_pipe_fd);
+		IGNORE_EINTR (close (slave_fifo_pipe_fd));
 		slave_fifo_pipe_fd = -1;
 	}
 
 	if (unixconn != NULL) {
 		gdm_connection_close (unixconn);
-		unlink (GDM_SUP_SOCKET);
+		IGNORE_EINTR (unlink (GDM_SUP_SOCKET));
 		unixconn = NULL;
 	}
 
 	closelog();
 
 	if (GdmPidFile != NULL)
-		unlink (GdmPidFile);
+		IGNORE_EINTR (unlink (GdmPidFile));
 }
 
 static gboolean
@@ -921,12 +1023,24 @@ deal_with_x_crashes (GdmDisplay *d)
 			"XKeepsCrashing script"),
 		      "deal_with_x_crashes");
 
-	    pid = gdm_fork_extra ();
+	    extra_process = pid = fork ();
+	    if (pid < 0)
+		    extra_process = 0;
 
 	    if (pid == 0) {
 		    char *argv[2];
 		    char *xlog = g_strconcat (GdmLogDir, "/", d->name, ".log", NULL);
 
+		    gdm_unset_signals ();
+
+		    /* Also make a new process group so that we may use
+		     * kill -(extra_process) to kill extra process and all it's
+		     * possible children */
+		    setsid ();
+
+		    if (GdmXdmcp)
+			    gdm_xdmcp_close ();
+
 		    closelog ();
 
 		    gdm_close_all_descriptors (0 /* from */, -1 /* except */, -1 /* except2 */);
@@ -958,14 +1072,36 @@ deal_with_x_crashes (GdmDisplay *d)
 		    ve_setenv ("TEXTDOMAIN", GETTEXT_PACKAGE, TRUE);
 		    ve_setenv ("TEXTDOMAINDIR", GNOMELOCALEDIR, TRUE);
 
-		    execv (argv[0], argv);
+		    IGNORE_EINTR (execv (argv[0], argv));
 	
 		    /* yaikes! */
 		    _exit (32);
 	    } else if (pid > 0) {
 		    int status;
 
-		    gdm_wait_for_extra (&status);
+		    if (extra_process > 1) {
+			    int ret;
+			    int killsignal = SIGTERM;
+			    int storeerrno;
+			    errno = 0;
+			    ret = waitpid (extra_process, &status, WNOHANG);
+			    do {
+				    /* wait for some signal, yes this is a race */
+				    if (ret <= 0)
+					    sleep (10);
+				    errno = 0;
+				    ret = waitpid (extra_process, &status, WNOHANG);
+				    storeerrno = errno;
+				    if (ret <= 0 &&
+					(ve_signal_was_notified (SIGTERM) ||
+					 ve_signal_was_notified (SIGINT) ||
+					 ve_signal_was_notified (SIGHUP))) {
+					    kill (-(extra_process), killsignal);
+					    killsignal = SIGKILL;
+				    }
+			    } while (ret == 0 || (ret < 0 && storeerrno == EINTR));
+		    }
+		    extra_process = 0;
 
 		    if (WIFEXITED (status) &&
 			WEXITSTATUS (status) == 0) {
@@ -1027,10 +1163,10 @@ suspend_machine (void)
 		/* Also make a new process group */
 		setsid ();
 
-		chdir ("/");
+		IGNORE_EINTR (chdir ("/"));
 
 		argv = ve_split (GdmSuspendReal);
-		execv (argv[0], argv);
+		IGNORE_EINTR (execv (argv[0], argv));
 		/* FIXME: what about fail */
 		_exit (1);
 	}
@@ -1041,8 +1177,8 @@ static void
 change_to_first_and_clear (gboolean reboot)
 {
 	gdm_change_vt (1);
-	close (1);
-	close (2);
+	IGNORE_EINTR (close (1));
+	IGNORE_EINTR (close (2));
 	open ("/dev/tty1", O_WRONLY);
 	open ("/dev/tty1", O_WRONLY);
 	/* yet again lazy */
@@ -1078,14 +1214,17 @@ gdm_cleanup_children (void)
     if (WIFEXITED (exitstatus)) {
 	    status = WEXITSTATUS (exitstatus);
 	    crashed = FALSE;
+	    gdm_debug ("gdm_cleanup_children: child %d returned %d", pid, status);
     } else {
 	    status = EXIT_SUCCESS;
 	    crashed = TRUE;
+	    if (WIFSIGNALED (exitstatus))
+		    gdm_debug ("gdm_cleanup_children: child %d crashed of signal %d", pid,
+			       (int)WTERMSIG (exitstatus));
+	    else
+		    gdm_debug ("gdm_cleanup_children: child %d crashed", pid);
     }
 	
-    gdm_debug ("gdm_cleanup_children: child %d returned %d%s", pid, status,
-	       crashed ? " (child crashed)" : "");
-
     if (pid == extra_process) {
 	    /* an extra process died, yay! */
 	    extra_process = 0;
@@ -1117,7 +1256,7 @@ gdm_cleanup_children (void)
 	    d->servpid = 0;
 
 	    /* race avoider */
-	    sleep (1);
+	    gdm_sleep_no_signal (1);
     }
 
     /* null all these, they are not valid most definately */
@@ -1222,14 +1361,14 @@ start_autopsy:
 	gdm_info (_("Master rebooting..."));
 
 	gdm_final_cleanup ();
-	chdir ("/");
+	IGNORE_EINTR (chdir ("/"));
 
 #ifdef __linux__
 	change_to_first_and_clear (TRUE /* reboot */);
 #endif /* __linux */
 
 	argv = ve_split (GdmRebootReal);
-	execv (argv[0], argv);
+	IGNORE_EINTR (execv (argv[0], argv));
 
 	gdm_error (_("%s: Reboot failed: %s"), 
 		   "gdm_child_action", strerror (errno));
@@ -1242,14 +1381,14 @@ start_autopsy:
 	gdm_info (_("Master halting..."));
 
 	gdm_final_cleanup ();
-	chdir ("/");
+	IGNORE_EINTR (chdir ("/"));
 
 #ifdef __linux__
 	change_to_first_and_clear (FALSE /* reboot */);
 #endif /* __linux */
 
 	argv = ve_split (GdmHaltReal);
-	execv (argv[0], argv);
+	IGNORE_EINTR (execv (argv[0], argv));
 
 	gdm_error (_("%s: Halt failed: %s"),
 		   "gdm_child_action", strerror (errno));
@@ -1273,6 +1412,7 @@ start_autopsy:
 	break;
 
     case DISPLAY_XFAILED:       /* X sucks */
+	gdm_debug ("X failed!");
 	/* inform about error if needed */
 	if (d->socket_conn != NULL) {
 		GdmConnection *conn = d->socket_conn;
@@ -1320,7 +1460,7 @@ start_autopsy:
 
     case DISPLAY_REMANAGE:	/* Remanage display */
     default:
-	gdm_debug ("gdm_child_action: Slave process returned %d", status);
+	gdm_debug ("gdm_child_action: In remanage");
 
 	/* inform about error if needed */
 	if (d->socket_conn != NULL) {
@@ -1373,7 +1513,7 @@ gdm_restart_now (void)
 	gdm_info (_("GDM restarting ..."));
 	gdm_final_cleanup ();
 	gdm_restoreenv ();
-	execvp (stored_argv[0], stored_argv);
+	IGNORE_EINTR (execvp (stored_argv[0], stored_argv));
 	gdm_error (_("Failed to restart self"));
 	_exit (1);
 }
@@ -1487,8 +1627,8 @@ create_connections (void)
 						 &pipeconn,
 						 close_notify);
 	} else {
-		close (p[0]);
-		close (p[1]);
+		IGNORE_EINTR (close (p[0]));
+		IGNORE_EINTR (close (p[1]));
 		slave_fifo_pipe_fd = -1;
 	}
 
@@ -1509,6 +1649,8 @@ create_connections (void)
 struct poptOption options [] = {
 	{ "nodaemon", '\0', POPT_ARG_NONE | POPT_ARGFLAG_ONEDASH,
 	  &no_daemon, 0, N_("Do not fork into the background"), NULL },
+	{ "no-console", '\0', POPT_ARG_NONE,
+	  &no_console, 0, N_("No console (local) servers to be run"), NULL },
 	{ "preserve-ld-vars", '\0', POPT_ARG_NONE,
 	  &preserve_ld_vars, 0, N_("Preserve LD_* variables"), NULL },
         POPT_AUTOHELP
@@ -1554,7 +1696,7 @@ ensure_desc_012 (void)
 		/* Once we are up to 3, we're beyond stdin,
 		 * stdout and stderr */
 		if (fd >= 3) {
-			close (fd);
+			IGNORE_EINTR (close (fd));
 			break;
 		}
 	}
@@ -1578,11 +1720,21 @@ gdm_get_our_runlevel (void)
 #endif /* __linux__ */
 }
 
+/* initially if we get a TERM or INT we just want to die,
+   but we want to also kill an extra process if it exists */
+static void
+initial_term_int (int signal)
+{
+	if (extra_process > 1)
+		kill (-(extra_process), SIGTERM);
+	_exit (EXIT_FAILURE);
+}
+
 int 
 main (int argc, char *argv[])
 {
     sigset_t mask;
-    struct sigaction term, child;
+    struct sigaction sig, child;
     FILE *pf;
     poptContext ctx;
     int nextopt;
@@ -1612,10 +1764,11 @@ main (int argc, char *argv[])
 	/* do nothing */ ;
 
     if (nextopt != -1) {
-	    printf (_("Error on option %s: %s.\nRun '%s --help' to see a full list of available command line options.\n"),
-		    poptBadOption (ctx, 0),
-		    poptStrerror (nextopt),
-		    argv[0]);
+	    fprintf (stderr,
+		     _("Error on option %s: %s.\nRun '%s --help' to see a full list of available command line options.\n"),
+		     poptBadOption (ctx, 0),
+		     poptStrerror (nextopt),
+		     argv[0]);
 	    fflush (stdout);
 	    exit (1);
     }
@@ -1634,6 +1787,19 @@ main (int argc, char *argv[])
 	    gdm_charset = g_strdup (charset);
     }
 
+    /* initial TERM/INT handler */
+    sig.sa_handler = initial_term_int;
+    sig.sa_flags = SA_RESTART;
+    sigemptyset (&sig.sa_mask);
+
+    if (sigaction (SIGTERM, &sig, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),
+		  "main", "TERM", strerror (errno));
+
+    if (sigaction (SIGINT, &sig, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),
+		  "main", "INT", strerror (errno));
+
     /* get the name of the root user */
     gdm_root_user ();
 
@@ -1671,7 +1837,7 @@ main (int argc, char *argv[])
 	    fclose (pf);
 	}
 
-	chdir (GdmServAuthDir);
+	IGNORE_EINTR (chdir (GdmServAuthDir));
 	umask (022);
     }
     else
@@ -1684,25 +1850,25 @@ main (int argc, char *argv[])
     ve_signal_add (SIGHUP, mainloop_sig_callback, NULL);
     ve_signal_add (SIGUSR1, mainloop_sig_callback, NULL);
     
-    term.sa_handler = ve_signal_notify;
-    term.sa_flags = SA_RESTART;
-    sigemptyset (&term.sa_mask);
+    sig.sa_handler = ve_signal_notify;
+    sig.sa_flags = SA_RESTART;
+    sigemptyset (&sig.sa_mask);
 
-    if (sigaction (SIGTERM, &term, NULL) < 0) 
-	gdm_fail (_("%s: Error setting up TERM signal handler"),
-		  "gdm_main");
+    if (sigaction (SIGTERM, &sig, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),
+		  "main", "TERM", strerror (errno));
 
-    if (sigaction (SIGINT, &term, NULL) < 0) 
-	gdm_fail (_("%s: Error setting up INT signal handler"),
-		  "gdm_main");
+    if (sigaction (SIGINT, &sig, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),
+		  "main", "INT", strerror (errno));
 
-    if (sigaction (SIGHUP, &term, NULL) < 0) 
-	gdm_fail (_("%s: Error setting up HUP signal handler"),
-		  "gdm_main");
+    if (sigaction (SIGHUP, &sig, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),
+		  "main", "HUP", strerror (errno));
 
-    if (sigaction (SIGUSR1, &term, NULL) < 0) 
-	gdm_fail (_("%s: Error setting up USR1 signal handler"),
-		  "gdm_main");
+    if (sigaction (SIGUSR1, &sig, NULL) < 0) 
+	gdm_fail (_("%s: Error setting up %s signal handler: %s"),
+		  "main", "USR1", strerror (errno));
 
     child.sa_handler = ve_signal_notify;
     child.sa_flags = SA_RESTART|SA_NOCLDSTOP;
@@ -1720,6 +1886,9 @@ main (int argc, char *argv[])
     sigaddset (&mask, SIGUSR1);
     sigprocmask (SIG_UNBLOCK, &mask, NULL);
 
+    gdm_signal_ignore (SIGUSR2);
+    gdm_signal_ignore (SIGPIPE);
+
     gdm_debug ("gdm_main: Here we go...");
 
     /* Init XDMCP if applicable */
@@ -1830,20 +1999,23 @@ send_slave_ack (GdmDisplay *d, const char *resp)
 			char not[2];
 			not[0] = GDM_SLAVE_NOTIFY_ACK;
 			not[1] = '\n';
-			write (d->master_notify_fd, not, 2);
+			IGNORE_EINTR (write (d->master_notify_fd, not, 2));
 		} else {
 			char *not = g_strdup_printf ("%c%s\n",
 						     GDM_SLAVE_NOTIFY_ACK,
 						     resp);
-			write (d->master_notify_fd, not, strlen (not));
+			IGNORE_EINTR (write (d->master_notify_fd, not, strlen (not)));
 			g_free (not);
 		}
 	}
-	gdm_sigchld_block_push ();
 	if (d->slavepid > 1) {
 		kill (d->slavepid, SIGUSR2);
+		/* now yield the CPU as the other process has more
+		   useful work to do then we do */
+#if defined(_POSIX_PRIORITY_SCHEDULING) && defined(HAVE_SCHED_YIELD)
+		sched_yield ();
+#endif
 	}
-	gdm_sigchld_block_pop ();
 }
 
 static void
@@ -1853,14 +2025,17 @@ send_slave_command (GdmDisplay *d, const char *command)
 		char *cmd = g_strdup_printf ("%c%s\n",
 					     GDM_SLAVE_NOTIFY_COMMAND,
 					     command);
-		write (d->master_notify_fd, cmd, strlen (cmd));
+		IGNORE_EINTR (write (d->master_notify_fd, cmd, strlen (cmd)));
 		g_free (cmd);
 	}
-	gdm_sigchld_block_push ();
 	if (d->slavepid > 1) {
 		kill (d->slavepid, SIGUSR2);
+		/* now yield the CPU as the other process has more
+		   useful work to do then we do */
+#if defined(_POSIX_PRIORITY_SCHEDULING) && defined(HAVE_SCHED_YIELD)
+		sched_yield ();
+#endif
 	}
-	gdm_sigchld_block_pop ();
 }
 
 
@@ -2265,10 +2440,10 @@ gdm_handle_message (GdmConnection *conn, const char *msg, gpointer data)
 		GSList *li;
 		for (li = displays; li != NULL; li = li->next) {
 			GdmDisplay *d = li->data;
-			gdm_sigchld_block_push ();
 			if (d->greetpid > 1)
 				kill (d->greetpid, SIGHUP);
-			gdm_sigchld_block_pop ();
+			else if (d->chooserpid > 1)
+				kill (d->chooserpid, SIGHUP);
 		}
 	} else if (strncmp (msg, GDM_SOP_WRITE_X_SERVERS " ",
 		            strlen (GDM_SOP_WRITE_X_SERVERS " ")) == 0) {
@@ -2437,7 +2612,7 @@ dehex_cookie (const char *cookie, int *len)
 	for (i = 0, p = cookie;
 	     *p != '\0' && *(p+1) != '\0';
 	     i++, p += 2) {
-		int num;
+		unsigned int num;
 		if (sscanf (p, "%02x", &num) != 1) {
 			g_free (bcookie);
 			return NULL;
@@ -2509,6 +2684,7 @@ handle_flexi_server (GdmConnection *conn, int type, const char *server,
 
 	if (type == TYPE_FLEXI_XNEST) {
 		struct stat s;
+		int r;
 		gboolean authorized = TRUE;
 
 		seteuid (xnest_uid);
@@ -2517,7 +2693,8 @@ handle_flexi_server (GdmConnection *conn, int type, const char *server,
 		g_assert (xnest_disp != NULL);
 		g_assert (xnest_cookie != NULL);
 
-		if (stat (xnest_auth_file, &s) < 0)
+		IGNORE_EINTR (r = stat (xnest_auth_file, &s));
+		if (r < 0)
 			authorized = FALSE;
 		if (authorized &&
 		    /* if readable or writable by group or others,
@@ -2659,10 +2836,8 @@ notify_displays_int (const char *key, int val)
 			gdm_fdprintf (disp->master_notify_fd,
 				      "%c%s %d\n",
 				      GDM_SLAVE_NOTIFY_KEY, key, val);
-			gdm_sigchld_block_push ();
 			if (disp->slavepid > 1)
 				kill (disp->slavepid, SIGUSR2);
-			gdm_sigchld_block_pop ();
 		}
 	}
 }
@@ -2677,10 +2852,8 @@ notify_displays_string (const char *key, const char *val)
 			gdm_fdprintf (disp->master_notify_fd,
 				      "%c%s %s\n",
 				      GDM_SLAVE_NOTIFY_KEY, key, val);
-			gdm_sigchld_block_push ();
 			if (disp->slavepid > 1)
 				kill (disp->slavepid, SIGUSR2);
-			gdm_sigchld_block_pop ();
 		}
 	}
 }
@@ -2690,8 +2863,10 @@ update_config (const char *key)
 {
 	struct stat statbuf;
 	VeConfig *cfg;
+	int r;
 
-	if (stat (GDM_CONFIG_FILE, &statbuf) == -1) {
+	IGNORE_EINTR (r = stat (GDM_CONFIG_FILE, &statbuf));
+	if (r < 0) {
 		/* if the file didn't exist before either */
 		if (config_file_mtime == 0)
 			return TRUE;
