@@ -45,7 +45,7 @@
  * process exit codes */
 #define SERVER_TIMEOUT 2	/* Server didn't start */
 #define SERVER_DEAD 250		/* Server stopped */
-#define SERVER_STARTED 251	/* Server started but not ready for connections yet */
+#define SERVER_PENDING 251	/* Server started but not ready for connections yet */
 #define SERVER_RUNNING 252	/* Server running and ready for connections */
 #define SERVER_ABORT 253	/* Server failed badly. Suspending display. */
 
@@ -134,7 +134,7 @@ enum {
 #define GDM_KEY_HALT "daemon/HaltCommand=/usr/bin/poweroff;/sbin/poweroff;/sbin/shutdown -h now;/usr/sbin/shutdown -h now"
 #define GDM_KEY_INITDIR "daemon/DisplayInitDir=" EXPANDED_SYSCONFDIR "/gdm/Init"
 #define GDM_KEY_KILLIC "daemon/KillInitClients=true"
-#define GDM_KEY_LOGDIR "daemon/LogDir=" EXPANDED_AUTHDIR
+#define GDM_KEY_LOGDIR "daemon/LogDir=" EXPANDED_LOGDIR
 #define GDM_KEY_PATH "daemon/DefaultPath=/bin:/usr/bin:" X_CONF_PATH ":" EXPANDED_BINDIR
 #define GDM_KEY_PIDFILE "daemon/PidFile=/var/run/gdm.pid"
 #define GDM_KEY_POSTSESS "daemon/PostSessionScriptDir=" EXPANDED_SYSCONFDIR "/gdm/PostSession/"
@@ -672,6 +672,13 @@ enum {
 					  * from a local display we started */
 };
 
+#define IGNORE_EINTR(expr) \
+	do {		\
+		errno = 0;	\
+		expr;		\
+	} while (errno == EINTR);
+		
+
 #endif /* GDM_H */
 
 /* EOF */
