@@ -56,6 +56,12 @@
 
 static gboolean RUNNING_UNDER_GDM = FALSE;
 
+enum {
+	CHOOSER_LIST_ICON_COLUMN = 0,
+	CHOOSER_LIST_LABEL_COLUMN,
+	CHOOSER_LIST_HOST_COLUMN
+};
+
 typedef struct _GdmChooserHost GdmChooserHost;
 struct _GdmChooserHost {
     gchar *name;
@@ -63,13 +69,12 @@ struct _GdmChooserHost {
     struct in_addr ia;
     GdkPixbuf *picture;
     gboolean willing;
-    int pos;
 };
 
 
-static const gchar *scanning_message = N_("Please wait: scanning local network for XDMCP-enabled hosts...");
+static const gchar *scanning_message = N_("Please wait: scanning local network...");
 static const gchar *empty_network = N_("No serving hosts were found.");
-static const gchar *active_network = N_("Choose a host to connect to from the selection below.");
+static const gchar *active_network = N_("Choose a ho_st to connect to:");
 
 /* XDM chooser style stuff */
 static gchar *xdm_address = NULL;
@@ -78,10 +83,12 @@ static gint connection_type = 0;
 
 static void gdm_chooser_abort (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
 static void gdm_chooser_warn (const gchar *format, ...) G_GNUC_PRINTF (1, 2);
+static void set_background (void);
 
 /* Exported for glade */
 void gdm_chooser_cancel (void);
 void gdm_chooser_add_host (void);
+void gdm_chooser_add_entry_changed (void);
 void gdm_chooser_manage (GtkButton *button, gpointer data);
 void gdm_chooser_browser_select (GtkWidget *widget,
 				 gint selected,
@@ -106,6 +113,10 @@ static guint scan_time_handler = 0;
 static int ping_tries = PING_TRIES;
 static guint ping_try_handler = 0;
 
+/* set in the main function */
+static char **stored_argv = NULL;
+static int stored_argc = 0;
+
 /* Fixetyfix */
 int XdmcpReallocARRAY8 (ARRAY8Ptr array, int length);
 
@@ -136,6 +147,7 @@ static gchar *GdmHostIconDir;
 static gchar *GdmHostDefaultIcon;
 static gchar *GdmGtkRC;
 static gchar *GdmHosts;
+static gchar *GdmHostsOrig;
 static gboolean GdmBroadcast;
 static gboolean GdmAllowAdd;
 static gchar *GdmBackgroundColor;
@@ -147,15 +159,32 @@ enum {
 };
 
 static GladeXML *chooser_app;
-static GtkWidget *chooser, *manage, *rescan, *cancel, *add_entry;
+static GtkWidget *chooser, *manage, *rescan, *cancel, *add_entry, *add_button;
 static GtkWidget *status_label;
 
 static GIOChannel *channel;
 static GList *hosts = NULL;
 static GdkPixbuf *defhostimg;
 static GtkWidget *browser;
+static GtkTreeModel *browser_model;
 static GdmChooserHost *curhost;
 
+static gboolean
+find_host_in_list (GdmChooserHost *host, GtkTreeIter *iter)
+{
+	if (gtk_tree_model_get_iter_first (browser_model, iter)) {
+		do {
+			GdmChooserHost *lhost;
+			gtk_tree_model_get (browser_model, iter,
+					    CHOOSER_LIST_HOST_COLUMN, &lhost,
+					    -1);
+			if (lhost == host)
+				return TRUE;
+		} while (gtk_tree_model_iter_next (browser_model, iter));
+	}
+	return FALSE;
+}
+
 static void
 setup_cursor (GdkCursorType type)
 {
@@ -181,18 +210,6 @@ gdm_chooser_host_dispose (GdmChooserHost *host)
     g_free (host);
 }
 
-static gint 
-gdm_chooser_sort_func (gpointer d1, gpointer d2)
-{
-    GdmChooserHost *a = d1;
-    GdmChooserHost *b = d2;
-
-    if (!d1 || !d2)
-	return 0;
-
-    return strcmp (a->name, b->name);
-}
-
 static GdmChooserHost * 
 gdm_chooser_host_alloc (const char *hostname,
 			const char *description,
@@ -202,28 +219,14 @@ gdm_chooser_host_alloc (const char *hostname,
     GdmChooserHost *host;
     GdkPixbuf *img;
     gchar *hostimg;
-    GList *hostl;
 
     host = g_new0 (GdmChooserHost, 1);
-    host->pos = -1;
     host->name = g_strdup (hostname);
     host->desc = g_strdup (description);
     memcpy (&host->ia, ia, sizeof (struct in_addr));
     host->willing = willing;
 
-    hostl = g_list_find_custom (hosts,
-				host,
-				(GCompareFunc) gdm_chooser_sort_func);
-    /* replace */
-    if (hostl != NULL) {
-	    GdmChooserHost *old = hostl->data;
-	    hostl->data = host;
-	    gdm_chooser_host_dispose (old);
-    } else {
-	    hosts = g_list_insert_sorted (hosts, 
-					  host,
-					  (GCompareFunc) gdm_chooser_sort_func);
-    }
+    hosts = g_list_prepend (hosts, host);
     
     if ( ! willing)
 	    return host;
@@ -271,22 +274,46 @@ static void
 gdm_chooser_browser_add_host (GdmChooserHost *host)
 {
     if (host->willing) {
-	    /* FIXME: the \n doesn't actually propagate
-	     * since the icon list is a broken piece of horsedung */
-	    char *temp = g_strconcat (host->name, " \n",
-				      host->desc, NULL);
-	    host->pos =
-		    gnome_icon_list_append_pixbuf (GNOME_ICON_LIST (browser),
-						   host->picture,
-						   NULL /* icon_filename */,
-						   temp);
-	    g_free (temp);
+	    GtkTreeIter iter = {0};
+	    const char *addr = inet_ntoa (host->ia);
+	    char *label;
+	    char *name, *desc;
+
+	    name = g_markup_escape_text (host->name, -1);
+	    desc = g_markup_escape_text (host->desc, -1);
+
+	    if (strcmp (addr, host->name) == 0)
+		    label = g_strdup_printf ("<b>%s</b>\n%s",
+					     name,
+					     desc);
+	    else
+		    label = g_strdup_printf ("<b>%s</b> (%s)\n%s",
+					     name,
+					     addr,
+					     desc);
+
+	    g_free (name);
+	    g_free (desc);
+
+	    gtk_list_store_append (GTK_LIST_STORE (browser_model), &iter);
+	    gtk_list_store_set (GTK_LIST_STORE (browser_model), &iter,
+				CHOOSER_LIST_ICON_COLUMN, host->picture,
+				CHOOSER_LIST_LABEL_COLUMN, label,
+				CHOOSER_LIST_HOST_COLUMN, host,
+				-1);
+	    g_free (label);
 
 	    if (added_addr != NULL &&
 		memcmp (&host->ia, added_addr,
 			sizeof (struct in_addr)) == 0) {
-		    gnome_icon_list_select_icon (GNOME_ICON_LIST (browser),
-						 host->pos);
+		    GtkTreeSelection *selection;
+		    GtkTreePath *path = gtk_tree_model_get_path (browser_model, &iter);
+		    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (browser));
+		    gtk_tree_selection_select_iter (selection, &iter);
+		    gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (browser),
+						  path, NULL,
+						  FALSE, 0.0, 0.0);
+		    gtk_tree_path_free (path);
 		    gtk_widget_grab_focus (manage);
 	    }
 	    g_free (added_host);
@@ -355,6 +382,7 @@ gdm_chooser_decode_packet (GIOChannel   *source,
     gchar *status = NULL;
     ARRAY8 auth, host, stat;
     GdmChooserHost *gh;
+    int pipe_buf;
 
     if (! XdmcpFill (sockfd, &buf, (XdmcpNetaddr) &clnt_sa, &sa_len))
 	return TRUE;
@@ -375,7 +403,7 @@ gdm_chooser_decode_packet (GIOChannel   *source,
 	    if (! XdmcpReadARRAY8 (&buf, &stat))
 		    goto done;
 
-	    status = g_strndup (stat.data, stat.length);
+	    status = g_strndup (stat.data, MIN (stat.length, 256));
     } else if (header.opcode == UNWILLING) {
 	    /* immaterial, will not be shown */
 	    status = NULL;
@@ -390,10 +418,6 @@ gdm_chooser_decode_packet (GIOChannel   *source,
 
 	    hostname = (he && he->h_name) ? he->h_name : inet_ntoa (clnt_sa.sin_addr);
 
-	    /* We can't pipe hostnames larger than this */
-	    if (strlen (hostname)+1 > PIPE_BUF)
-		    goto done;
-
 	    hostname = g_strdup (hostname);
     } else {
 	    hostname = g_new0 (char, 1024);
@@ -403,6 +427,20 @@ gdm_chooser_decode_packet (GIOChannel   *source,
 	    }
     }
 
+#ifdef PIPE_BUF
+    pipe_buf = PIPE_BUF;
+#else
+    /* apparently Hurd doesn't have PIPE_BUF */
+    pipe_buf = fpathconf (1 /*stdout*/, _PC_PIPE_BUF);
+    /* could return -1 if no limit */
+#endif
+
+    /* We can't pipe hostnames larger than this */
+    if (pipe_buf > 0 && strlen (hostname)+1 > pipe_buf) {
+	    g_free (hostname);
+	    goto done;
+    }
+
     gh = gdm_host_known (&clnt_sa.sin_addr);
     if (gh == NULL) {
 	    gh = gdm_chooser_host_alloc (hostname,
@@ -435,16 +473,16 @@ gdm_chooser_decode_packet (GIOChannel   *source,
 		    g_source_remove (add_check_handler);
 	    add_check_handler = 0;
 
-	    dialog = gtk_message_dialog_new
+	    dialog = ve_hig_dialog_new
 		    (GTK_WINDOW (chooser) /* parent */,
 		     GTK_DIALOG_MODAL /* flags */,
 		     GTK_MESSAGE_ERROR,
-		     GTK_BUTTONS_CLOSE,
+		     GTK_BUTTONS_OK,
+		     _("Cannot connect to remote server"),
 		     _("The host \"%s\" is not willing "
 		       "to support a login session right now.  "
 		       "Please try again later."),
 		     added_host);
-	    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 
 	    if (RUNNING_UNDER_GDM)
 		    gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -545,9 +583,9 @@ chooser_scan_time_update (gpointer data)
 			break;
 	}
 	if (li != NULL /* something was found */) {
-		gtk_label_set_text (GTK_LABEL (status_label), _(active_network));
+		gtk_label_set_label (GTK_LABEL (status_label), _(active_network));
 	} else {
-		gtk_label_set_text (GTK_LABEL (status_label), _(empty_network));
+		gtk_label_set_label (GTK_LABEL (status_label), _(empty_network));
 	}
 	gtk_widget_set_sensitive (GTK_WIDGET (rescan), TRUE);
 	return FALSE;
@@ -608,10 +646,10 @@ gdm_chooser_xdmcp_discover (void)
 
     gtk_widget_set_sensitive (GTK_WIDGET (manage), FALSE);
     gtk_widget_set_sensitive (GTK_WIDGET (rescan), FALSE);
-    gnome_icon_list_clear (GNOME_ICON_LIST (browser));
+    gtk_list_store_clear (GTK_LIST_STORE (browser_model));
     gtk_widget_set_sensitive (GTK_WIDGET (browser), FALSE);
-    gtk_label_set_text (GTK_LABEL (status_label),
-			_(scanning_message));
+    gtk_label_set_label (GTK_LABEL (status_label),
+			 _(scanning_message));
 
     while (hl) {
 	gdm_chooser_host_dispose ((GdmChooserHost *) hl->data);
@@ -802,11 +840,12 @@ add_check (gpointer data)
 	if (added_addr != NULL) {
 		GtkWidget *dialog;
 
-		dialog = gtk_message_dialog_new
+		dialog = ve_hig_dialog_new
 			(GTK_WINDOW (chooser) /* parent */,
 			 GTK_DIALOG_MODAL /* flags */,
 			 GTK_MESSAGE_ERROR,
-			 GTK_BUTTONS_CLOSE,
+			 GTK_BUTTONS_OK,
+			 _("Did not receive response from server"),
 			 _("Did not receive any response from host \"%s\" "
 			   "in %d seconds.  Perhaps the host is not "
 			   "turned on, or is not willing to support a "
@@ -814,7 +853,6 @@ add_check (gpointer data)
 			   "later."),
 			 added_host,
 			 ADD_TIMEOUT / 1000);
-		gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 
 		if (RUNNING_UNDER_GDM)
 			gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -855,15 +893,15 @@ gdm_chooser_add_host (void)
 	} else {
 		GtkWidget *dialog;
 
-		dialog = gtk_message_dialog_new
+		dialog = ve_hig_dialog_new
 			(GTK_WINDOW (chooser) /* parent */,
 			 GTK_DIALOG_MODAL /* flags */,
 			 GTK_MESSAGE_ERROR,
-			 GTK_BUTTONS_CLOSE,
+			 GTK_BUTTONS_OK,
+			 _("Cannot find host"),
 			 _("I cannot find the host \"%s\", "
 			   "perhaps you have mistyped it."),
 			 name);
-		gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 
 		if (RUNNING_UNDER_GDM)
 			gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -887,9 +925,16 @@ gdm_chooser_add_host (void)
 
 	host = gdm_host_known (ia);
 	if (host != NULL) {
-		if (host->pos >= 0) {
-			gnome_icon_list_select_icon (GNOME_ICON_LIST (browser),
-						     host->pos);
+		GtkTreeIter iter = {0};
+		if (find_host_in_list (host, &iter)) {
+			GtkTreeSelection *selection;
+			GtkTreePath *path = gtk_tree_model_get_path (browser_model, &iter);
+			selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (browser));
+			gtk_tree_selection_select_iter (selection, &iter);
+			gtk_tree_view_scroll_to_cell (GTK_TREE_VIEW (browser),
+						      path, NULL,
+						      FALSE, 0.0, 0.0);
+			gtk_tree_path_free (path);
 			gtk_widget_grab_focus (manage);
 		} else {
 			/* hmm, probably not willing, ping the host then for
@@ -930,6 +975,15 @@ gdm_chooser_add_host (void)
 	gtk_entry_set_text (GTK_ENTRY (add_entry), "");
 }
 
+void
+gdm_chooser_add_entry_changed (void)
+{
+	const char *name;
+
+	name = gtk_entry_get_text (GTK_ENTRY (add_entry));
+	gtk_widget_set_sensitive (add_button, ! ve_string_empty (name));
+}
+
 void
 gdm_chooser_cancel (void)
 {
@@ -998,31 +1052,28 @@ gdm_chooser_warn (const gchar *format, ...)
 static void 
 gdm_chooser_parse_config (void)
 {
-    /* stupid, or is it?
-    struct stat unused;
-    if (stat (GDM_CONFIG_FILE, &unused) == -1)
-	gdm_chooser_abort (_("gdm_chooser_parse_config: No configuration file: %s. Aborting."), GDM_CONFIG_FILE);
-	*/
+    VeConfig *cfg;
 
-    gnome_config_push_prefix ("=" GDM_CONFIG_FILE "=/");
+    cfg = ve_config_get (GDM_CONFIG_FILE);
 
-    GdmXineramaScreen = gnome_config_get_int (GDM_KEY_XINERAMASCREEN);
-    GdmGtkRC = gnome_config_get_string (GDM_KEY_GTKRC);
-    GdmScanTime = gnome_config_get_int (GDM_KEY_SCAN);
-    GdmHostDefaultIcon = gnome_config_get_string (GDM_KEY_HOST);
-    GdmHostIconDir = gnome_config_get_string (GDM_KEY_HOSTDIR);
-    GdmIconMaxWidth = gnome_config_get_int (GDM_KEY_ICONWIDTH);
-    GdmIconMaxHeight = gnome_config_get_int (GDM_KEY_ICONHEIGHT);
-    GdmDebug = gnome_config_get_bool (GDM_KEY_DEBUG);
+    GdmXineramaScreen = ve_config_get_int (cfg, GDM_KEY_XINERAMASCREEN);
+    GdmGtkRC = ve_config_get_string (cfg, GDM_KEY_GTKRC);
+    GdmScanTime = ve_config_get_int (cfg, GDM_KEY_SCAN);
+    GdmHostDefaultIcon = ve_config_get_string (cfg, GDM_KEY_HOST);
+    GdmHostIconDir = ve_config_get_string (cfg, GDM_KEY_HOSTDIR);
+    GdmIconMaxWidth = ve_config_get_int (cfg, GDM_KEY_ICONWIDTH);
+    GdmIconMaxHeight = ve_config_get_int (cfg, GDM_KEY_ICONHEIGHT);
+    GdmDebug = ve_config_get_bool (cfg, GDM_KEY_DEBUG);
 
-    GdmBackgroundColor = gnome_config_get_string (GDM_KEY_BACKGROUNDCOLOR);
-    GdmBackgroundType = gnome_config_get_int (GDM_KEY_BACKGROUNDTYPE);
+    GdmBackgroundColor = ve_config_get_string (cfg, GDM_KEY_BACKGROUNDCOLOR);
+    GdmBackgroundType = ve_config_get_int (cfg, GDM_KEY_BACKGROUNDTYPE);
 
-    GdmAllowAdd = gnome_config_get_bool (GDM_KEY_ALLOWADD);
+    GdmAllowAdd = ve_config_get_bool (cfg, GDM_KEY_ALLOWADD);
 
     /* note that command line arguments will prevail over these */
-    GdmHosts = gnome_config_get_string (GDM_KEY_HOSTS);
-    GdmBroadcast = gnome_config_get_bool (GDM_KEY_BROADCAST);
+    GdmHosts = ve_config_get_string (cfg, GDM_KEY_HOSTS);
+    GdmHostsOrig = g_strdup (GdmHosts);
+    GdmBroadcast = ve_config_get_bool (cfg, GDM_KEY_BROADCAST);
     /* if broadcasting, then append BROADCAST to hosts */
     if (GdmBroadcast) {
 	    if (ve_string_empty (GdmHosts)) {
@@ -1038,47 +1089,31 @@ gdm_chooser_parse_config (void)
     if (GdmScanTime < 1) GdmScanTime = 1;
     if (GdmIconMaxWidth < 0) GdmIconMaxWidth = 128;
     if (GdmIconMaxHeight < 0) GdmIconMaxHeight = 128;
-
-    gnome_config_pop_prefix();
 }
 
-static GdmChooserHost *
-gdm_nth_willing_host (int n)
+static void
+host_selected (GtkTreeSelection *selection, gpointer data)
 {
-	GList *li;
-	int i;
+	GtkTreeModel *tm = NULL;
+	GtkTreeIter iter = {0};
 
-	i = 0;
-	for (li = hosts; li != NULL; li = li->next) {
-		GdmChooserHost *host = li->data;
-		if (host->willing) {
-			if (i == n)
-				return host;
-			i++;
-		}
-	}
-	return NULL;
-}
-
-
-void
-gdm_chooser_browser_select (GtkWidget *widget, gint selected, GdkEvent *event)
-{
-	curhost = gdm_nth_willing_host (selected);
-	gtk_widget_set_sensitive (manage, TRUE);
-
-	if (event != NULL &&
-	    event->type == GDK_2BUTTON_PRESS) {
-		gdm_chooser_manage (NULL, NULL);
-	}
-}
-
-
-void
-gdm_chooser_browser_unselect (GtkWidget *widget, gint selected, GdkEvent *event)
-{
 	curhost = NULL;
-	gtk_widget_set_sensitive (manage, FALSE);
+
+	if (gtk_tree_selection_get_selected (selection, &tm, &iter)) {
+		gtk_tree_model_get (tm, &iter, CHOOSER_LIST_HOST_COLUMN,
+				    &curhost, -1);
+	}
+
+	gtk_widget_set_sensitive (manage, curhost != NULL);
+}
+
+static void
+row_activated (GtkTreeView *tree_view,
+	       GtkTreePath *path,
+	       GtkTreeViewColumn *column)
+{
+	if (curhost != NULL)
+		gdm_chooser_manage (NULL, NULL);
 }
 
 void
@@ -1086,6 +1121,7 @@ display_chooser_information (void)
 {
 	GtkWidget *dialog;
 
+	/* How to get HIG compliance? */
 	dialog = gtk_message_dialog_new
 		(GTK_WINDOW (chooser) /* parent */,
 		 GTK_DIALOG_MODAL /* flags */,
@@ -1114,12 +1150,14 @@ gdm_chooser_gui_init (void)
 {
 	int width;
 	int height;
+	GtkTreeSelection *selection;
+	GtkTreeViewColumn *column;
 
 	glade_helper_add_glade_directory (GDM_GLADE_DIR);
 	glade_helper_search_gnome_dirs (FALSE);
 
     /* Enable theme */
-    if (GdmGtkRC)
+    if (RUNNING_UNDER_GDM && GdmGtkRC)
 	gtk_rc_parse (GdmGtkRC);
 
     /* Load default host image */
@@ -1154,14 +1192,49 @@ gdm_chooser_gui_init (void)
 				     GTK_TYPE_LABEL);
     add_entry = glade_helper_get (chooser_app, "add_entry",
 				  GTK_TYPE_ENTRY);
+    add_button = glade_helper_get (chooser_app, "add_button",
+				   GTK_TYPE_BUTTON);
 
     browser = glade_helper_get (chooser_app, "chooser_iconlist",
-				GNOME_TYPE_ICON_LIST);
-    gnome_icon_list_freeze (GNOME_ICON_LIST (browser));
-    gnome_icon_list_set_separators (GNOME_ICON_LIST (browser), " /-_.\n");
-    gnome_icon_list_set_icon_width (GNOME_ICON_LIST (browser), GdmIconMaxWidth + 20);
-    gnome_icon_list_set_icon_border (GNOME_ICON_LIST (browser), 2);
-    gnome_icon_list_thaw (GNOME_ICON_LIST (browser));
+				GTK_TYPE_TREE_VIEW);
+
+    gtk_dialog_set_has_separator (GTK_DIALOG (chooser), FALSE);
+
+    gtk_tree_view_set_rules_hint (GTK_TREE_VIEW (browser), TRUE);
+
+    selection = gtk_tree_view_get_selection (GTK_TREE_VIEW (browser));
+    gtk_tree_selection_set_mode (selection, GTK_SELECTION_SINGLE);
+
+    g_signal_connect (selection, "changed",
+		      G_CALLBACK (host_selected),
+		      NULL);
+    g_signal_connect (browser, "row_activated",
+		      G_CALLBACK (row_activated),
+		      NULL);
+
+    browser_model = (GtkTreeModel *)gtk_list_store_new (3,
+							GDK_TYPE_PIXBUF,
+							G_TYPE_STRING,
+							G_TYPE_POINTER);
+    gtk_tree_view_set_model (GTK_TREE_VIEW (browser), browser_model);
+    column = gtk_tree_view_column_new_with_attributes
+	    ("Icon",
+	     gtk_cell_renderer_pixbuf_new (),
+	     "pixbuf", CHOOSER_LIST_ICON_COLUMN,
+	     NULL);
+    gtk_tree_view_append_column (GTK_TREE_VIEW (browser), column);
+
+    column = gtk_tree_view_column_new_with_attributes
+	    ("Hostname",
+	     gtk_cell_renderer_text_new (),
+	     "markup", CHOOSER_LIST_LABEL_COLUMN,
+	     NULL);
+    gtk_tree_view_append_column (GTK_TREE_VIEW (browser), column);
+
+    gtk_tree_sortable_set_sort_column_id (GTK_TREE_SORTABLE (browser_model),
+					  CHOOSER_LIST_LABEL_COLUMN,
+					  GTK_SORT_ASCENDING);
+
 
     if ( ! GdmAllowAdd) {
 	    GtkWidget *w = glade_helper_get (chooser_app, "add_hbox",
@@ -1195,25 +1268,112 @@ gdm_chooser_gui_init (void)
     }
 }
 
+static gboolean
+string_same (VeConfig *config, const char *cur, const char *key)
+{
+	char *val = ve_config_get_string (config, key);
+	if (strcmp (ve_sure_string (cur), ve_sure_string (val)) == 0) {
+		g_free (val);
+		return TRUE;
+	} else {
+		g_free (val);
+		return FALSE;
+	}
+}
+
+static gboolean
+bool_same (VeConfig *config, gboolean cur, const char *key)
+{
+	gboolean val = ve_config_get_bool (config, key);
+	if (ve_bool_equal (cur, val)) {
+		return TRUE;
+	} else {
+		return FALSE;
+	}
+}
+
+static gboolean
+int_same (VeConfig *config, int cur, const char *key)
+{
+	int val = ve_config_get_int (config, key);
+	if (cur == val) {
+		return TRUE;
+	} else {
+		return FALSE;
+	}
+}
+
+
+static gboolean
+gdm_reread_config (int sig, gpointer data)
+{
+	VeConfig *config;
+	/* reparse config stuff here.  At least ones we care about */
+
+	config = ve_config_get (GDM_CONFIG_FILE);
+
+	/* FIXME: The following is evil, we should update on the fly rather
+	 * then just restarting */
+	/* Also we may not need to check ALL those keys but just a few */
+	if ( ! string_same (config, GdmHostsOrig, GDM_KEY_HOSTS) ||
+	     ! string_same (config, GdmGtkRC, GDM_KEY_GTKRC) ||
+	     ! string_same (config, GdmHostDefaultIcon, GDM_KEY_HOST) ||
+	     ! string_same (config, GdmHostIconDir, GDM_KEY_HOSTDIR) ||
+	     ! int_same (config,
+			 GdmXineramaScreen, GDM_KEY_XINERAMASCREEN) ||
+	     ! int_same (config, GdmIconMaxWidth, GDM_KEY_ICONWIDTH) ||
+	     ! int_same (config, GdmIconMaxHeight, GDM_KEY_ICONHEIGHT) ||
+	     ! int_same (config, GdmScanTime, GDM_KEY_SCAN) ||
+	     ! bool_same (config, GdmDebug, GDM_KEY_DEBUG)) {
+		if (RUNNING_UNDER_GDM) {
+			/* Set busy cursor */
+			setup_cursor (GDK_WATCH);
+
+			gdm_wm_save_wm_order ();
+		}
+
+		/* we don't need to tell the slave that we're restarting
+		   it doesn't care about our state.  Unlike with the greeter */
+		execvp (stored_argv[0], stored_argv);
+		_exit (DISPLAY_REMANAGE);
+	}
+
+	/* we only use the color and do it for all types except NONE */
+	if ( ! string_same (config, GdmBackgroundColor, GDM_KEY_BACKGROUNDCOLOR) ||
+	     ! int_same (config, GdmBackgroundType, GDM_KEY_BACKGROUNDTYPE)) {
+		set_background ();
+	}
+
+	return TRUE;
+}
+
 
 static void
 gdm_chooser_signals_init (void)
 {
     struct sigaction hup;
+    struct sigaction term;
     sigset_t mask;
 
-    hup.sa_handler = (void *) gdm_chooser_cancel;
+    ve_signal_add (SIGHUP, gdm_reread_config, NULL);
+
+    hup.sa_handler = ve_signal_notify;
     hup.sa_flags = 0;
-    sigemptyset (&hup.sa_mask);
+    sigemptyset(&hup.sa_mask);
+    sigaddset (&hup.sa_mask, SIGCHLD);
+
+    term.sa_handler = (void *) gdm_chooser_cancel;
+    term.sa_flags = 0;
+    sigemptyset (&term.sa_mask);
 
     if (sigaction (SIGHUP, &hup, NULL) < 0) 
-        gdm_chooser_abort (_("gdm_signals_init: Error setting up HUP signal handler"));
+        gdm_chooser_abort (_("%s: Error setting up %s signal handler: %s"), "gdm_signals_init", "HUP", strerror (errno));
 
-    if (sigaction (SIGINT, &hup, NULL) < 0) 
-        gdm_chooser_abort (_("gdm_signals_init: Error setting up INT signal handler"));
+    if (sigaction (SIGINT, &term, NULL) < 0) 
+        gdm_chooser_abort (_("%s: Error setting up %s signal handler: %s"), "gdm_signals_init", "INT", strerror (errno));
 
-    if (sigaction (SIGTERM, &hup, NULL) < 0) 
-        gdm_chooser_abort (_("gdm_signals_init: Error setting up TERM signal handler"));
+    if (sigaction (SIGTERM, &term, NULL) < 0) 
+        gdm_chooser_abort (_("%s: Error setting up %s signal handler: %s"), "gdm_signals_init", "TERM", strerror (errno));
 
     sigfillset (&mask);
     sigdelset (&mask, SIGTERM);
@@ -1303,6 +1463,13 @@ main (int argc, char *argv[])
     poptContext ctx;
     int nextopt;
     const char *gdm_version;
+    int i;
+
+    stored_argv = g_new0 (char *, argc + 1);
+    for (i = 0; i < argc; i++)
+	    stored_argv[i] = g_strdup (argv[i]);
+    stored_argv[i] = NULL;
+    stored_argc = argc;
 
     if (g_getenv ("RUNNING_UNDER_GDM") != NULL)
 	    RUNNING_UNDER_GDM = TRUE;
@@ -1350,16 +1517,16 @@ main (int argc, char *argv[])
 
 	    gdm_wm_focus_new_windows (TRUE);
 
-	    dialog = gtk_message_dialog_new (NULL /* parent */,
-					     GTK_DIALOG_MODAL /* flags */,
-					     GTK_MESSAGE_ERROR,
-					     GTK_BUTTONS_OK,
-					     _("The chooser version (%s) does not match the daemon "
-					       "version (%s).\n"
-					       "You have probably just upgraded gdm.\n"
-					       "Please restart the gdm daemon or reboot the computer."),
-					     VERSION, gdm_version);
-	    gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
+	    dialog = ve_hig_dialog_new (NULL /* parent */,
+					GTK_DIALOG_MODAL /* flags */,
+					GTK_MESSAGE_ERROR,
+					GTK_BUTTONS_OK,
+					_("Cannot run chooser"),
+					_("The chooser version (%s) does not match the daemon "
+					  "version (%s).  "
+					  "You have probably just upgraded gdm.  "
+					  "Please restart the gdm daemon or reboot the computer."),
+					VERSION, gdm_version);
 
 	    gtk_widget_show_all (dialog);
 	    gdm_wm_center_window (GTK_WINDOW (dialog));
@@ -1417,12 +1584,16 @@ main (int argc, char *argv[])
 	    gdm_wm_focus_window (GDK_WINDOW_XWINDOW (chooser->window));
     }
 
-    if (RUNNING_UNDER_GDM)
-	    setup_cursor (GDK_LEFT_PTR);
-
     if (GdmAllowAdd)
 	    gtk_widget_grab_focus (add_entry);
 
+    gdm_chooser_add_entry_changed ();
+
+    if (RUNNING_UNDER_GDM) {
+	    gdm_wm_restore_wm_order ();
+	    setup_cursor (GDK_LEFT_PTR);
+    }
+
     gtk_main();
 
     exit (EXIT_SUCCESS);
