@@ -22,6 +22,7 @@
 #include <libgnome/libgnome.h>
 #include <libgnomeui/libgnomeui.h>
 #include <sys/types.h>
+#include <signal.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 #include <errno.h>
@@ -57,23 +58,21 @@ gdm_check (void)
 	if (pid <= 1 ||
 	    (kill (pid, 0) < 0 &&
 	     errno != EPERM)) {
-		dialog = gtk_message_dialog_new
+		dialog = ve_hig_dialog_new
 			(NULL /* parent */,
 			 GTK_DIALOG_MODAL /* flags */,
 			 GTK_MESSAGE_WARNING,
 			 GTK_BUTTONS_OK,
-			 "foo");
-		gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
-		gtk_label_set_markup
-			 (GTK_LABEL (GTK_MESSAGE_DIALOG (dialog)->label),
-			  _("<b>GDM (The GNOME Display Manager) "
-			    "is not running.</b>\n\n"
-			    "You might in fact be using a different "
-			    "display manager, such as KDM "
-			    "(KDE Display Manager or xdm).\n"
-			    "If you still wish to use this feature, "
-			    "either start GDM your self or ask your "
-			    "system administrator to start GDM."));
+			 _("GDM (The GNOME Display Manager) "
+			   "is not running."),
+			 "%s\n%s",
+			 _("You might in fact be using a different "
+			   "display manager, such as KDM "
+			   "(KDE Display Manager or xdm)."),
+			 _("If you still wish to use this feature, "
+			   "either start GDM yourself or ask your "
+			   "system administrator to start GDM."));
+
 		gtk_widget_show_all (dialog);
 		gtk_dialog_run (GTK_DIALOG (dialog));
 		gtk_widget_destroy (dialog);
@@ -98,6 +97,8 @@ main (int argc, char *argv[])
 	GtkWidget *dialog;
 	GtkWidget *photo;
 	gboolean face_browser;
+	char *greeter;
+	char *remotegreeter;
 	int max_size;
 	char *last_pix;
 
@@ -115,29 +116,40 @@ main (int argc, char *argv[])
 	gnome_config_push_prefix ("=" GDM_CONFIG_FILE "=/");
 	face_browser = gnome_config_get_bool (GDM_KEY_BROWSER);
 	max_size = gnome_config_get_int (GDM_KEY_MAXFILE);
+	greeter = gnome_config_get_string (GDM_KEY_GREETER);
+	remotegreeter = gnome_config_get_string (GDM_KEY_REMOTEGREETER);
 	gnome_config_pop_prefix ();
 
 	if ( ! gdm_check ()) {
 		return 1;
 	}
 
-	if ( ! face_browser) {
+	/* HACK */
+	/* only warn if gdmlogin is set for both local and remote greeter,
+	 * the graphical greeter does a different setup thingie for
+	 * the face browser and it would be hard to figure out here ... */
+	if ( ! face_browser &&
+	     strstr (greeter, "gdmlogin") != NULL &&
+	     strstr (remotegreeter, "gdmlogin") != NULL) {
 		GtkWidget *d;
-		d = gtk_message_dialog_new (NULL /* parent */,
-					    GTK_DIALOG_MODAL /* flags */,
-					    GTK_MESSAGE_WARNING,
-					    GTK_BUTTONS_OK,
-					    _("The face browser is not "
-					      "configured,\nplease ask your "
-					      "system administrator to enable "
-					      "it\nin the GDM configurator "
-					      "program."));
-		gtk_dialog_set_has_separator (GTK_DIALOG (d), FALSE);
+		d = ve_hig_dialog_new (NULL /* parent */,
+				       GTK_DIALOG_MODAL /* flags */,
+				       GTK_MESSAGE_WARNING,
+				       GTK_BUTTONS_OK,
+				       _("The face browser is not "
+					 "configured"),
+				       "%s",
+				       _("The face browser is not configured in the "
+					 "GDM configuration.  Please ask your "
+					 "system administrator to enable "
+					 "it in the GDM configurator "
+					 "program."));
+
 		gtk_dialog_run (GTK_DIALOG (d));
 		gtk_widget_destroy (d);
 	}
 
-	dialog = gtk_dialog_new_with_buttons (_("Select a photo"),
+	dialog = gtk_dialog_new_with_buttons (_("Login Photo"),
 					      NULL /* parent */,
 					      0 /* flags */,
 					      GTK_STOCK_CANCEL,
@@ -145,6 +157,7 @@ main (int argc, char *argv[])
 					      GTK_STOCK_OK,
 					      GTK_RESPONSE_OK,
 					      NULL);
+	gtk_dialog_set_has_separator (GTK_DIALOG (dialog), FALSE);
 	gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox),
 			    gtk_label_new (_("Select a photograph to show "
 					     "in the facebrowser:")),
@@ -180,34 +193,33 @@ main (int argc, char *argv[])
 		if (ve_string_empty (pixmap) ||
 		    stat (pixmap, &s) < 0) {
 			GtkWidget *d;
-			d = gtk_message_dialog_new (NULL /* parent */,
-						    GTK_DIALOG_MODAL /* flags */,
-						    GTK_MESSAGE_WARNING,
-						    GTK_BUTTONS_OK,
-						    _("No picture selected."));
-			gtk_dialog_set_has_separator (GTK_DIALOG (d), FALSE);
+			d = ve_hig_dialog_new (NULL /* parent */,
+					       GTK_DIALOG_MODAL /* flags */,
+					       GTK_MESSAGE_WARNING,
+					       GTK_BUTTONS_OK,
+					       _("No picture selected."),
+					       /* avoid warning */ "%s", "");
 			gtk_dialog_run (GTK_DIALOG (d));
 			gtk_widget_destroy (d);
 		} else if (s.st_size > max_size) {
 			GtkWidget *d;
-			d = gtk_message_dialog_new (NULL /* parent */,
-						    GTK_DIALOG_MODAL /* flags */,
-						    GTK_MESSAGE_WARNING,
-						    GTK_BUTTONS_OK,
-						    _("The picture is too large and "
-						      "the system administrator\n"
-						      "disallowed pictures larger "
-						      "then %d bytes to\n"
-						      "show in the face browser"),
-						    max_size);
-			gtk_dialog_set_has_separator (GTK_DIALOG (d), FALSE);
+			d = ve_hig_dialog_new (NULL /* parent */,
+					       GTK_DIALOG_MODAL /* flags */,
+					       GTK_MESSAGE_WARNING,
+					       GTK_BUTTONS_OK,
+					       _("Picture is too large"),
+					       _("The system administrator "
+						 "disallowed pictures larger "
+						 "than %d bytes to "
+						 "show in the face browser"),
+					       max_size);
 			gtk_dialog_run (GTK_DIALOG (d));
 			gtk_widget_destroy (d);
 		} else {
 			char buf[4096];
 			size_t size;
 			char *photofile = g_strconcat (g_get_home_dir (),
-						       "/.gnome2/photo",
+						       "/.face",
 						       NULL);
 			char *cfg_file = g_strconcat (g_get_home_dir (),
 						      "/.gnome2/gdm",
@@ -217,15 +229,15 @@ main (int argc, char *argv[])
 			fdsrc = open (pixmap, O_RDONLY);
 			if (fdsrc < 0) {
 				GtkWidget *d;
-				d = gtk_message_dialog_new (NULL /* parent */,
-							    GTK_DIALOG_MODAL /* flags */,
-							    GTK_MESSAGE_ERROR,
-							    GTK_BUTTONS_OK,
-							    _("File %s cannot be open for "
-							      "reading\nError: %s"),
-							    pixmap,
-							    g_strerror (errno));
-				gtk_dialog_set_has_separator (GTK_DIALOG (d), FALSE);
+				d = ve_hig_dialog_new (NULL /* parent */,
+						       GTK_DIALOG_MODAL /* flags */,
+						       GTK_MESSAGE_ERROR,
+						       GTK_BUTTONS_OK,
+						       _("Cannot open file"),
+						       _("File %s cannot be open for "
+							 "reading\nError: %s"),
+						       pixmap,
+						       g_strerror (errno));
 				gtk_dialog_run (GTK_DIALOG (d));
 				gtk_widget_destroy (d);
 				g_free (cfg_file);
@@ -236,15 +248,15 @@ main (int argc, char *argv[])
 			fddest = open (photofile, O_WRONLY | O_CREAT);
 			if (fddest < 0) {
 				GtkWidget *d;
-				d = gtk_message_dialog_new (NULL /* parent */,
-							    GTK_DIALOG_MODAL /* flags */,
-							    GTK_MESSAGE_ERROR,
-							    GTK_BUTTONS_OK,
-							    _("File %s cannot be open for "
-							      "writing\nError: %s"),
-							    photofile,
-							    g_strerror (errno));
-				gtk_dialog_set_has_separator (GTK_DIALOG (d), FALSE);
+				d = ve_hig_dialog_new (NULL /* parent */,
+						       GTK_DIALOG_MODAL /* flags */,
+						       GTK_MESSAGE_ERROR,
+						       GTK_BUTTONS_OK,
+						       _("Cannot open file"),
+						       _("File %s cannot be open for "
+							 "writing\nError: %s"),
+						       photofile,
+						       g_strerror (errno));
 				gtk_dialog_run (GTK_DIALOG (d));
 				gtk_widget_destroy (d);
 				g_free (cfg_file);
