@@ -1,5 +1,5 @@
-#ifndef __GREETER_EVENTS_H__
-#define __GREETER_EVENTS_H__
+#ifndef GREETER_EVENTS_H
+#define GREETER_EVENTS_H
 
 typedef void  (*ActionFunc) (GreeterItemInfo *info,
 			     gpointer         user_data);
@@ -12,7 +12,6 @@ void greeter_item_register_action_callback (char            *id,
 					    ActionFunc       func,
 					    gpointer         user_data);
 
-				       
+void greeter_item_run_action_callback (const char *id);
 
-
-#endif /* __GREETER_EVENTS_H__ */
+#endif /* GREETER_EVENTS_H */
