@@ -19,11 +19,6 @@
 #ifndef MISC_H
 #define MISC_H
 
-#include <gnome.h>
-
-void entry_set_red (GtkWidget *w, gboolean state);
-int strcasecmp_no_locale (const char *s1, const char *s2);
-
 #endif /* MISC_H */
 
 /* EOF */
